# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import tempfile
import unittest

import torch
from atek.data_loaders.atek_wds_dataloader import load_atek_wds_dataset
from atek.data_preprocess.atek_wds_writer import AtekWdsWriter
from atek.data_preprocess.sample_builders.obb_sample_builder import ObbSampleBuilder
from atek.data_preprocess.subsampling_lib.temporal_subsampler import (
    CameraTemporalSubsampler,
)
from atek.util.tensor_utils import check_dicts_same_w_tensors

from omegaconf import OmegaConf


# test data paths
TEST_DIR_PATH = os.path.join(os.getenv("TEST_FOLDER"))
CONFIG_PATH = os.getenv("CONFIG_PATH")
CATEGORY_MAPPING_PATH = os.getenv("CATEGORY_MAPPING_PATH")


class LoadAtekWdsDatasetTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.temp_dir_object = tempfile.TemporaryDirectory()
        self.output_wds_path = self.temp_dir_object.name

    def _preprocess_data(self) -> None:
        """
        A helper function to preprocess some ATEK data and write to WDS.
        """
        # From raw Aria dataset to WDS
        conf = OmegaConf.load(CONFIG_PATH)

        sample_builder = ObbSampleBuilder(
            conf=conf.processors,
            vrs_file=os.path.join(TEST_DIR_PATH, "test_ADT_unit_test_sequence.vrs"),
            sequence_name="test",
            mps_files={
                "mps_closedloop_traj_file": os.path.join(
                    TEST_DIR_PATH, "test_ADT_trajectory.csv"
                ),
            },
            gt_files={
                "obb3_file": os.path.join(TEST_DIR_PATH, "test_3d_bounding_box.csv"),
                "obb3_traj_file": os.path.join(
                    TEST_DIR_PATH, "test_3d_bounding_box_traj.csv"
                ),
                "obb2_file": os.path.join(TEST_DIR_PATH, "test_2d_bounding_box.csv"),
                "instance_json_file": os.path.join(
                    TEST_DIR_PATH, "test_instances.json"
                ),
                "category_mapping_file": CATEGORY_MAPPING_PATH,
            },
        )

        subsampler = CameraTemporalSubsampler(
            vrs_file=os.path.join(TEST_DIR_PATH, "test_ADT_unit_test_sequence.vrs"),
            conf=conf.camera_temporal_subsampler,
        )

        atek_wds_writer = AtekWdsWriter(
            output_path=self.output_wds_path,
            conf=conf.wds_writer,
        )

        # will be used for comparison after loading data
        self.expected_flattened_samples = []

        for i in range(subsampler.get_total_num_samples()):
            timestamps_ns = subsampler.get_timestamps_by_sample_index(i)

            for t in timestamps_ns:
                sample = sample_builder.get_sample_by_timestamp_ns(t)
                if sample is not None:
                    atek_wds_writer.add_sample(data_sample=sample)
                    self.expected_flattened_samples.append(sample.to_flatten_dict())
        atek_wds_writer.close()

    def test_atek_native_round_trip(self) -> None:
        # Preprocess data and create WDS files
        self._preprocess_data()

        # check number of tars created
        output_wds_names = os.listdir(self.output_wds_path)
        self.assertEqual(len(output_wds_names), 1)  # only 1 tar file created

        # Load tars back into ATEK
        tar_list = [os.path.join(self.output_wds_path, f) for f in output_wds_names]
        unbatched_dataset = load_atek_wds_dataset(
            tar_list, batch_size=None, repeat_flag=False
        )

        for expected_sample, sample in zip(
            self.expected_flattened_samples, unbatched_dataset
        ):
            # Check GT content
            self.assertTrue(
                check_dicts_same_w_tensors(
                    expected_sample["gt_data"], sample["gt_data"], atol=1e-3
                )
            )

            # Check sequence name
            self.assertEqual(expected_sample["sequence_name"], sample["sequence_name"])

            # Check other data content, note that expected_sample's keys have file extensions, like "gtdata.json",
            # while ATEK loaded samples won't have extensions.
            for key_wo_extension, expected_val in expected_sample.items():
                # skip certain fields
                if "image" in key_wo_extension or "gt_data" in key_wo_extension:
                    continue

                self.assertTrue(key_wo_extension in sample)
                sample_val = sample[key_wo_extension]

                if isinstance(expected_val, torch.Tensor):
                    self.assertTrue(
                        torch.allclose(
                            expected_val.to(torch.float32),
                            sample_val.to(torch.float32),
                            atol=1e-3,
                        )
                    )
                else:
                    self.assertEqual(expected_val, sample_val)

    def test_atek_default_collation(self) -> None:
        # Preprocess data and create WDS files
        self._preprocess_data()

        # check number of tars created
        output_wds_names = os.listdir(self.output_wds_path)
        self.assertEqual(len(output_wds_names), 1)  # only 1 tar file created

        # Load tars back into ATEK
        tar_list = [os.path.join(self.output_wds_path, f) for f in output_wds_names]
        batch_size = 2  # TODO: maybe support a unit test with a larger dataset
        batched_dataset = load_atek_wds_dataset(
            tar_list, batch_size=batch_size, repeat_flag=False
        )

        # Check batched data tensor shape
        batched_sample = next(iter(batched_dataset))
        for key, val in batched_sample.items():
            if key in ["__key__", "__url__"]:
                continue

            if isinstance(val, torch.Tensor):
                self.assertEqual(val.shape[0], batch_size)
            elif isinstance(val, list):
                self.assertEqual(len(val), batch_size)

    # This is like a DTOR
    def tearDown(self):
        # Explicitly cleanup the temporary directory
        self.temp_dir_object.cleanup()
        print("Temporary directory deleted")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

import torch

from atek.data_preprocess.processors.aria_camera_processor import AriaCameraProcessor
from omegaconf import OmegaConf


# test data paths
TEST_VRS_PATH = os.path.join(
    os.getenv("TEST_FOLDER"), "test_ADT_unit_test_sequence.vrs"
)
CONFIG_PATH = os.getenv("CONFIG_PATH")


class AriaCameraProcessorTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def _single_case_test_get_image_data(
        self, camera_processor, gt_timestamps_ns, gt_frame_id, gt_image_shape
    ):
        """
        A helper function to perform a test on a single camera processor
        """
        # Expect to success, only 20ns away
        query_timestamps = gt_timestamps_ns + 20
        maybe_result = camera_processor.get_image_data_by_timestamps_ns(
            timestamps_ns=query_timestamps.tolist()
        )
        self.assertTrue(maybe_result is not None)

        self.assertTrue(torch.allclose(maybe_result.frame_ids, gt_frame_id, atol=1))
        self.assertTrue(
            torch.allclose(maybe_result.capture_timestamps_ns, gt_timestamps_ns, atol=1)
        )
        self.assertEqual(maybe_result.images.shape, gt_image_shape)

    def test_get_image_data(self) -> None:
        conf = OmegaConf.load(CONFIG_PATH)

        # Test for Slam left
        slam_camera_processor = AriaCameraProcessor(
            TEST_VRS_PATH, conf.processors.slam_left
        )
        self._single_case_test_get_image_data(
            slam_camera_processor,
            gt_timestamps_ns=torch.tensor(
                [87551270894700, 87551337550700], dtype=torch.int64
            ),
            gt_frame_id=torch.tensor([1335, 1337], dtype=torch.int64),
            gt_image_shape=torch.Size([2, 1, 480, 640]),
        )

        # Test for rotated SLAM left
        slam_conf_2 = OmegaConf.merge(
            conf.processors.slam_left, {"rotate_image_cw90deg": True}
        )
        slam_camera_processor_2 = AriaCameraProcessor(TEST_VRS_PATH, slam_conf_2)
        self._single_case_test_get_image_data(
            slam_camera_processor_2,
            gt_timestamps_ns=torch.tensor([87551337550700], dtype=torch.int64),
            gt_frame_id=torch.tensor([1337], dtype=torch.int64),
            gt_image_shape=torch.Size([1, 1, 640, 480]),
        )

        # Test for RGB camera
        rgb_conf = OmegaConf.merge(
            conf.processors.rgb, {"target_camera_resolution": [512, 512]}
        )
        rgb_camera_processor = AriaCameraProcessor(TEST_VRS_PATH, rgb_conf)
        self._single_case_test_get_image_data(
            rgb_camera_processor,
            gt_timestamps_ns=torch.tensor([87551337447475], dtype=torch.int64),
            gt_frame_id=torch.tensor([1335], dtype=torch.int64),
            gt_image_shape=torch.Size([1, 3, 512, 512]),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

import torch

from atek.data_preprocess.atek_data_sample import (
    AtekDataSample,
    MpsOnlineCalibData,
    MpsSemiDensePointData,
    MpsTrajData,
    MultiFrameCameraData,
)


class MultiFrameCameraDataTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_flatten_to_dict(self) -> None:
        data = MultiFrameCameraData()
        data.images = torch.randn(4, 1, 10, 10)
        data.frame_ids = torch.tensor([1, 2, 3, 4])

        data.camera_label = "test-camera"
        data.projection_params = torch.randn(20, 1)

        data_dict = data.to_flatten_dict()
        expected_keys = [
            "mfcd#test-camera+images",
            "mfcd#test-camera+frame_ids",
            "mfcd#test-camera+camera_label",
            "mfcd#test-camera+projection_params",
        ]
        self.assertCountEqual(data_dict.keys(), expected_keys)


class TestMpsSemiDensePointData(unittest.TestCase):
    def test_to_flatten_dict(self):
        # Create an instance of MpsSemiDensePointData with sample data
        point_data = MpsSemiDensePointData(
            points_world=[
                torch.tensor([[1, 2, 3], [4, 5, 6]]),
                torch.tensor([[7, 8, 9]]),
            ],
            points_inv_dist_std=[torch.tensor([0.1, 0.2]), torch.tensor([0.3])],
            points_dist_std=[torch.tensor([0.5, 0.4]), torch.tensor([0.3])],
            capture_timestamps_ns=torch.tensor([1, 2]),
        )

        # Call the method
        flatten_dict = point_data.to_flatten_dict()

        # Check the keys in the flatten_dict
        expected_keys = [
            "msdpd#points_world",
            "msdpd#points_inv_dist_std",
            "msdpd#points_dist_std",
            "msdpd#capture_timestamps_ns",
        ]
        self.assertCountEqual(flatten_dict.keys(), expected_keys)

        # Check value for both timestamps
        for i_timestamp in range(2):
            self.assertTrue(
                torch.allclose(
                    flatten_dict["msdpd#points_world"][i_timestamp],
                    point_data.points_world[i_timestamp],
                    atol=1e-5,
                )
            )

            self.assertTrue(
                torch.allclose(
                    flatten_dict["msdpd#points_inv_dist_std"][i_timestamp],
                    point_data.points_inv_dist_std[i_timestamp],
                    atol=1e-5,
                )
            )

            self.assertTrue(
                torch.allclose(
                    flatten_dict["msdpd#points_dist_std"][i_timestamp],
                    point_data.points_dist_std[i_timestamp],
                    atol=1e-5,
                )
            )


class AtekDataSampleTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_flatten_to_dict(self) -> None:
        rgb_data = MultiFrameCameraData(
            images=torch.randn(4, 3, 10, 10),
            camera_label="camera-rgb",
            projection_params=torch.randn(20, 1),
        )

        slam_left_data = MultiFrameCameraData(
            images=torch.randn(4, 1, 10, 12),
            camera_label="camera-slam-left",
            frame_ids=torch.tensor([1, 2, 3, 4]),
        )

        mps_traj_data = MpsTrajData(
            Ts_World_Device=torch.randn(1, 3, 4),
            gravity_in_world=torch.tensor([0, 0, -9.81]),
        )

        mps_online_calib_data = MpsOnlineCalibData(
            capture_timestamps_ns=torch.tensor([14580434854000]),
            utc_timestamps_ns=torch.tensor([0], dtype=torch.int64),
            projection_params=torch.tensor(
                [
                    [
                        [
                            2.4138e02,
                            3.1623e02,
                            2.3668e02,
                            -2.7658e-02,
                            1.0371e-01,
                            -7.3428e-02,
                            1.2858e-02,
                            1.3407e-03,
                            -4.5717e-04,
                            4.4453e-04,
                            -1.9325e-03,
                            1.5349e-04,
                            2.6054e-04,
                            3.3760e-03,
                            1.3415e-04,
                        ],
                        [
                            2.4104e02,
                            3.1769e02,
                            2.3818e02,
                            -2.4638e-02,
                            9.5806e-02,
                            -6.5053e-02,
                            9.1013e-03,
                            1.9200e-03,
                            -4.4897e-04,
                            1.5518e-03,
                            -5.5009e-04,
                            -8.7365e-04,
                            2.1657e-04,
                            1.1816e-03,
                            1.9690e-04,
                        ],
                        [
                            1.2219e03,
                            1.4627e03,
                            1.4659e03,
                            4.0604e-01,
                            -4.8995e-01,
                            1.7457e-01,
                            1.1330e00,
                            -1.7016e00,
                            6.5116e-01,
                            6.2115e-04,
                            1.9322e-05,
                            -1.4855e-05,
                            2.6012e-04,
                            -6.5821e-04,
                            3.7614e-05,
                        ],
                    ]
                ],
                dtype=torch.float64,
            ),
            ts_device_camera=torch.tensor(
                [
                    [
                        [
                            [9.9998e-01, 5.5086e-04, -5.9830e-03, 4.3290e-04],
                            [-6.2809e-04, 9.9992e-01, -1.2914e-02, -9.0162e-05],
                            [5.9753e-03, 1.2917e-02, 9.9990e-01, -4.8937e-05],
                            [0.0000e00, 0.0000e00, 0.0000e00, 1.0000e00],
                        ],
                        [
                            [9.9845e-01, -3.4170e-02, 4.3828e-02, 6.0829e-03],
                            [5.0779e-02, 2.4046e-01, -9.6933e-01, -1.1159e-01],
                            [2.2584e-02, 9.7006e-01, 2.4182e-01, -8.7817e-02],
                            [0.0000e00, 0.0000e00, 0.0000e00, 1.0000e00],
                        ],
                        [
                            [9.9339e-01, -5.1506e-02, 1.0257e-01, -4.2807e-03],
                            [1.0364e-01, 7.8649e-01, -6.0884e-01, -1.1842e-02],
                            [-4.9313e-02, 6.1545e-01, 7.8663e-01, -5.1140e-03],
                            [0.0000e00, 0.0000e00, 0.0000e00, 1.0000e00],
                        ],
                    ]
                ]
            ),
        )

        gt_data = {"key_1": "value_1", "key_2": "value_2"}

        data = AtekDataSample(
            camera_rgb=rgb_data,
            camera_slam_left=slam_left_data,
            mps_traj_data=mps_traj_data,
            gt_data=gt_data,
            mps_online_calib_data=mps_online_calib_data,
        )

        data_dict = data.to_flatten_dict()
        expected_keys = [
            "mfcd#camera-rgb+images",
            "mfcd#camera-rgb+camera_label",
            "mfcd#camera-rgb+projection_params",
            "mfcd#camera-slam-left+images",
            "mfcd#camera-slam-left+camera_label",
            "mfcd#camera-slam-left+frame_ids",
            "mtd#ts_world_device",
            "mtd#gravity_in_world",
            "gt_data",
            "mocd#capture_timestamps_ns",
            "mocd#utc_timestamps_ns",
            "mocd#projection_params",
            "mocd#ts_device_camera",
        ]
        # for testing only
        print(f"------ debug: data dict keys: {data_dict.keys()} ------")
        self.assertCountEqual(data_dict.keys(), expected_keys)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

import torch

from atek.data_preprocess.processors.aria_camera_processor import AriaCameraProcessor
from atek.data_preprocess.processors.depth_image_processor import DepthImageProcessor
from omegaconf import OmegaConf
from torchvision.transforms import InterpolationMode


# test data paths
TEST_FOLDER = os.getenv("TEST_FOLDER")
CONFIG_PATH = os.getenv("CONFIG_PATH")


class DepthImageProcessorTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def _single_case_test_get_depth_data(
        self, camera_processor, gt_timestamps_ns, gt_frame_id, gt_image_shape
    ):
        """
        A helper function to perform a test on a single camera processor
        """
        # Expect to success, only 20ns away
        query_timestamps = gt_timestamps_ns + 20
        maybe_result = camera_processor.get_depth_data_by_timestamps_ns(
            timestamps_ns=query_timestamps.tolist()
        )
        self.assertTrue(maybe_result is not None)

        self.assertTrue(torch.allclose(maybe_result.frame_ids, gt_frame_id, atol=1))
        self.assertTrue(
            torch.allclose(maybe_result.capture_timestamps_ns, gt_timestamps_ns, atol=1)
        )
        self.assertEqual(maybe_result.images.shape, gt_image_shape)

    def test_get_depth_image_data(self) -> None:
        conf = OmegaConf.load(CONFIG_PATH)

        # Obtain image transformations from AriaCameraProcessor
        rgb_conf = conf.processors.rgb
        rgb_conf = OmegaConf.merge(rgb_conf, {"undistort_to_linear_camera": True})
        rgb_conf = OmegaConf.merge(rgb_conf, {"target_camera_resolution": [704, 704]})
        rgb_conf = OmegaConf.merge(rgb_conf, {"rotate_image_cw90deg": True})
        rgb_camera_processor = AriaCameraProcessor(
            video_vrs=os.path.join(TEST_FOLDER, "test_ADT_unit_test_sequence.vrs"),
            conf=rgb_conf,
        )
        depth_image_transform = rgb_camera_processor.get_image_transform(
            rescale_interpolation=InterpolationMode.NEAREST
        )

        # Construct DepthImageProcessor
        depth_conf = conf.processors.rgb_depth
        # replace type id with exact stream id, and turn on depth conversion, for ADT data.
        depth_conf.pop("depth_stream_type_id")
        depth_conf = OmegaConf.merge(
            depth_conf, {"depth_stream_id": "345-1", "convert_zdepth_to_distance": True}
        )
        rgb_depth_processor = DepthImageProcessor(
            depth_vrs=os.path.join(TEST_FOLDER, "test_ADT_depth_rgb_only.vrs"),
            image_transform=depth_image_transform,
            depth_camera_label="camera-rgb-depth",
            depth_camera_calib=rgb_camera_processor.get_final_camera_calib(),
            conf=depth_conf,
        )

        self._single_case_test_get_depth_data(
            rgb_depth_processor,
            gt_timestamps_ns=torch.tensor(
                [87551204237999, 87551237566000], dtype=torch.int64
            ),
            gt_frame_id=torch.tensor([1, 2], dtype=torch.int64),
            gt_image_shape=torch.Size([2, 1, 704, 704]),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from dataclasses import fields
from typing import Dict, List, Optional

import torch

from atek.data_preprocess.atek_data_sample import AtekDataSample, MpsTrajData
from atek.data_preprocess.processors.aria_camera_processor import AriaCameraProcessor
from atek.data_preprocess.processors.depth_image_processor import DepthImageProcessor
from atek.data_preprocess.processors.efm_gt_processor import EfmGtProcessor
from atek.data_preprocess.processors.mps_semidense_processor import (
    MpsSemiDenseProcessor,
)
from atek.data_preprocess.processors.mps_traj_processor import MpsTrajProcessor
from omegaconf.omegaconf import DictConfig
from torchvision.transforms import InterpolationMode


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class EfmSampleBuilder:
    """
    A customized sample builder for EFM, which includes both 3D obb detection task and surface reconstruction task.

    Models using this sample builder:
    - EFM
    """

    def __init__(
        self,
        conf: DictConfig,
        vrs_file: str,
        sequence_name: str,
        mps_files: Dict[str, str],
        gt_files: Dict[str, str],
        depth_vrs_file: str,
    ) -> None:
        """
        Initializes the EfmSampleBuilder with necessary configuration and file paths.

        Args:
            conf (DictConfig): Configuration object containing settings for various processors.
            vrs_file (str): Path to the main Aria VRS file used for video and sensor data.
            mps_files (Dict[str, str]): Dictionary mapping keys to file paths for MPS data, including:
                "mps_closedloop_traj_file": Path to the closed-loop trajectory file.
                "mps_semidense_points_file": Path to the global semidense points file.
                "mps_semidense_observations_file": Path to the observations of semidense points, indicating which points are observable by which camera, at each timestamp.
            gt_files (Dict[str, str]): Dictionary mapping keys to file paths for ground truth data, including:
                "obb3_file": Path to the 3D object bounding box file.
                "obb3_traj_file": Path to the 3D object bounding box trajectory file.
                "obb2_file": Path to the 2D object bounding box file.
                "instance_json_file": Path to the JSON file containing object instance information.
            depth_vrs_file [Optional]: Path to the depth VRS file. This is required for the surface recon task.
        """
        self.conf = conf

        self.vrs_file = vrs_file
        self.depth_vrs_file = depth_vrs_file
        self.sequence_name = sequence_name

        self.processors = self._add_processors_from_conf(
            conf, vrs_file, mps_files, gt_files
        )

    def _add_processors_from_conf(
        self,
        conf: DictConfig,
        vrs_file: str,
        mps_files: Dict[str, str],
        gt_files: Dict[str, str],
    ):
        """
        This function creates a dict of processors from the config file.
        """
        processors = {}
        # camera processors
        camera_conf_list = [
            conf.rgb,
            conf.slam_left,
            conf.slam_right,
        ]
        selected_camera_label_to_stream_ids = {}
        for camera_conf in camera_conf_list:
            if camera_conf.selected:
                cam_processor = AriaCameraProcessor(vrs_file, camera_conf)
                processors[camera_conf.sensor_label] = cam_processor
                selected_camera_label_to_stream_ids[camera_conf.sensor_label] = (
                    cam_processor.get_stream_id()
                )

        # MPS processors
        if "mps_traj" in conf and conf.mps_traj.selected:
            processors["mps_traj"] = MpsTrajProcessor(
                mps_closedloop_traj_file=mps_files["mps_closedloop_traj_file"],
                conf=conf.mps_traj,
            )

        if "mps_semidense" in conf and conf.mps_semidense.selected:
            processors["mps_semidense"] = MpsSemiDenseProcessor(
                mps_semidense_points_file=mps_files["mps_semidense_points_file"],
                mps_semidense_observations_file=mps_files[
                    "mps_semidense_observations_file"
                ],
                conf=conf.mps_semidense,
            )

        # Depth processor
        if (
            "rgb_depth" in conf
            and conf.rgb_depth.selected
            and os.path.exists(self.depth_vrs_file)
        ):
            assert (
                self.depth_vrs_file != ""
            ), "need to specify depth vrs file to use depth processor"

            # Obtain image transformations from rgb AriaCameraProcessor, where interpolation needs to be exactly set to NEAREST
            assert (
                "camera-rgb" in processors
            ), "rgb_depth depends on camera_rgb processor to obtain camera calibration"
            depth_image_transform = processors["camera-rgb"].get_image_transform(
                rescale_interpolation=InterpolationMode.NEAREST
            )
            depth_camera_calib = processors["camera-rgb"].get_final_camera_calib()

            processors["rgb_depth"] = DepthImageProcessor(
                depth_vrs=self.depth_vrs_file,
                image_transform=depth_image_transform,
                depth_camera_calib=depth_camera_calib,
                depth_camera_label="camera-rgb-depth",
                conf=conf.rgb_depth,
            )

        if len(gt_files) > 0 and "efm_gt" in conf and conf.efm_gt.selected:
            processors["efm_gt"] = EfmGtProcessor(
                obb3_file_path=gt_files["obb3_file"],
                obb3_traj_file_path=gt_files["obb3_traj_file"],
                obb2_file_path=gt_files["obb2_file"],
                instance_json_file_path=gt_files["instance_json_file"],
                category_mapping_file_path=gt_files.get(
                    "category_mapping_file", None
                ),  # this file is optional
                camera_label_to_stream_ids=selected_camera_label_to_stream_ids,
                conf=conf.efm_gt,
            )

        return processors

    def _check_timestamps_are_consistent(
        self, sample: AtekDataSample, timestamps_ns: List[int]
    ) -> bool:
        """
        Check if all timestamps in the sample are consistent.
        """
        num_timestamps = len(timestamps_ns)
        for data_field in fields(sample):
            name = data_field.name
            data_field_value = getattr(sample, name)
            if hasattr(data_field_value, "capture_timestamps_ns"):
                time = getattr(data_field_value, "capture_timestamps_ns")
                if len(time) != num_timestamps:
                    logger.warning(
                        f"Timestamps in {name} does not have full count of {num_timestamps}!"
                    )
                    return False

            # Check for gt
            if name == "gt_data" and "efm_gt" in data_field_value:
                if len(data_field_value["efm_gt"]) != num_timestamps:
                    logger.warning(
                        f"Timestamps in {name} does not have full count of {num_timestamps}!"
                    )
                    return False

        return True

    def get_sample_by_timestamps_ns(
        self, timestamps_ns: List[int]
    ) -> Optional[AtekDataSample]:
        sample = AtekDataSample()

        # First assign sequence name
        sample.sequence_name = self.sequence_name

        for processor_label, processor in self.processors.items():
            if isinstance(processor, AriaCameraProcessor):
                # ========================================
                # Aria camera sensor data
                # ========================================
                sample_camera_data = processor.get_image_data_by_timestamps_ns(
                    timestamps_ns=timestamps_ns
                )
                # Skip if no image data is available
                if sample_camera_data is None:
                    logger.warning(
                        f"Querying camera for {timestamps_ns} on processor {processor_label} has returned None, skipping this sample."
                    )
                    return None

                # Fill calibration data
                sample_camera_data.camera_label = processor_label
                sample_camera_data.origin_camera_label = processor.get_origin_label()
                final_camera_calib = processor.get_final_camera_calib()
                sample_camera_data.T_Device_Camera = torch.from_numpy(
                    final_camera_calib.get_transform_device_camera().to_matrix3x4()
                )
                sample_camera_data.camera_model_name = str(
                    final_camera_calib.model_name()
                )
                sample_camera_data.projection_params = torch.from_numpy(
                    final_camera_calib.projection_params()
                )
                sample_camera_data.camera_valid_radius = torch.tensor(
                    [final_camera_calib.get_valid_radius()], dtype=torch.float32
                )

                setattr(
                    sample,
                    processor_label.replace(
                        "-", "_"
                    ),  # field name in AtekDataSample is same as sensor label, apart from a `-`->`_` conversion
                    sample_camera_data,
                )

            # ========================================
            # MPS traj data
            # ========================================
            elif isinstance(processor, MpsTrajProcessor):
                maybe_mps_traj_data = processor.get_closed_loop_pose_by_timestamps_ns(
                    timestamps_ns
                )
                if maybe_mps_traj_data is None:
                    logger.warning(
                        f"Querying MPS traj for {timestamps_ns} has returned None, skipping this sample."
                    )
                    return None

                # Fill MPS traj data into sample
                sample.mps_traj_data = MpsTrajData(
                    Ts_World_Device=maybe_mps_traj_data[0],
                    capture_timestamps_ns=maybe_mps_traj_data[1],
                    gravity_in_world=maybe_mps_traj_data[2],
                )

            # =======================================
            # MPS SemiDense data
            # =======================================
            elif isinstance(processor, MpsSemiDenseProcessor):
                maybe_mps_semidense_data = (
                    processor.get_semidense_points_by_timestamps_ns(timestamps_ns)
                )
                if maybe_mps_semidense_data is None:
                    logger.warning(
                        f"Querying MPS SemiDense data for {timestamps_ns} has returned None, skipping this sample."
                    )
                    return None

                # Fill MPS SemiDense data into sample
                sample.mps_semidense_point_data = maybe_mps_semidense_data

            # =======================================
            # RGB Depth data
            # =======================================
            elif isinstance(processor, DepthImageProcessor):
                maybe_depth_data = processor.get_depth_data_by_timestamps_ns(
                    timestamps_ns
                )
                if maybe_depth_data is None:
                    logger.warning(
                        f"Querying Depth data for {timestamps_ns} has returned None, skipping this sample."
                    )
                    return None

                # Fill depth data into sample
                sample.camera_rgb_depth = maybe_depth_data

            # ========================================
            # GT data
            # ========================================
            elif isinstance(processor, EfmGtProcessor):
                maybe_gt_data = processor.get_gt_by_timestamp_list_ns(timestamps_ns)
                if maybe_gt_data is None:
                    logger.warning(
                        f"Querying GT data for {timestamps_ns} has returned None, skipping this sample."
                    )
                    return None
                sample.gt_data["efm_gt"] = maybe_gt_data

            else:
                raise ValueError(
                    f"Unimplemented processor class {processor.__name__} in SampleBuilder! "
                )

        # only return data of common timestamps among all sub-data types
        if not self._check_timestamps_are_consistent(sample, timestamps_ns):
            logger.warning(
                "Timestamps in sample are not consistent, skipping this sample."
            )
            return None

        return sample



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Callable, Dict, Optional, Tuple

import numpy as np
import torch

from atek.util.file_io_utils import load_category_mapping_from_csv

from omegaconf.omegaconf import DictConfig
from projectaria_tools.core.calibration import CameraCalibration

from projectaria_tools.projects.adt import (
    AriaDigitalTwinDataPaths,
    AriaDigitalTwinDataProvider,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ATEK_OTHER_CATETORY_ID: int = (
    0  # 0 is reserved for other categories in ATEK object taxonomy
)


class Obb2GtProcessor:
    """
    A Ground truth (GT) processor class for Object bounding box 2D (OBB2) data, used for 2D object detection ML task
    """

    def __init__(
        self,
        obb2_file_path: str,
        instance_json_file_path: str,
        category_mapping_file_path: Optional[str],
        camera_label_to_stream_ids: Dict[str, str],
        camera_label_to_pixel_transforms: Optional[Dict[str, Callable]],
        camera_label_to_calib: Optional[Dict[str, CameraCalibration]],
        conf: DictConfig,
    ) -> None:
        """
        Initializes the Obb2GtProcessor object with necessary file paths and configurations.

        Args:
            obb2_file_path (str): The file path to the OBB2 data file in ADT format.
            instance_json_file_path (str): The file path to the JSON file containing instance data in ADT format
            category_mapping_file_path (Optional[str]): Path to the CSV file mapping category names to IDs, in the format of:
                {
                    $KEY_TO_MAP： [“cat_name”, category_id],
                    ...
                },
                where "KEY_TO_MAP" is one of strings of {"prototype_name", "category"}, set through conf.category_mapping_field_name
            camera_label_to_stream_ids (Dict[str, str]): A dictionary mapping camera labels to stream IDs.
            camera_label_to_pixel_transforms (Optional[Dict[str, Callable]]): A dictionary mapping camera labels to pixel transformation functions, if available.
            camera_label_to_calib (Optional[Dict[str, CameraCalibration]]): A dictionary mapping camera labels to camera calibration data, if available.

            conf (DictConfig): A configuration object containing settings for the processor, example yaml:
                ```
                category_mapping_field_name: "prototype_name" # {prototype_name (for ADT), category (for ASE)}
                bbox2d_num_samples_on_edge: 10                 # Number of samples on each edge of 2D bbox in distortion
                tolerance_ns: 1000000                          # Tolerance in nanoseconds for timestamp validation
                ```
        """
        self.conf = conf

        # Create a ADT data provider object, which will be used to load 3D BBOX GT data
        data_paths = AriaDigitalTwinDataPaths()
        data_paths.boundingboxes_2d_filepath = obb2_file_path
        data_paths.instances_filepath = instance_json_file_path

        self.adt_gt_provider = AriaDigitalTwinDataProvider(data_paths)

        self.camera_label_to_stream_ids = camera_label_to_stream_ids
        self.camera_label_to_pixel_transforms = camera_label_to_pixel_transforms
        self.camera_label_to_calibs = camera_label_to_calib

        self.category_mapping = (
            None
            if category_mapping_file_path is None
            else load_category_mapping_from_csv(category_mapping_file_path)
        )

    def _obtain_obj_category_info(self, instance_id: int) -> Tuple[str, int]:
        """
        Retrieves the category name and ID for a given instance ID.
        Args:
            instance_id (int): The instance ID for which to retrieve category information.
        Returns:
            Tuple[str, int]: A tuple containing the category name and category ID.
        """
        instance_info = self.adt_gt_provider.get_instance_info_by_id(instance_id)

        if not self.category_mapping:
            # If no category mapping is provided, we use the original category name.
            category_name = instance_info.category
            category_id = instance_info.category_uid
        else:
            # Query the mapping field from instance
            key_to_map = getattr(
                instance_info, self.conf.category_mapping_field_name, None
            )
            if not key_to_map:
                raise ValueError(
                    f'Unsupported instance field to map: {self.conf.category_mapping_field_name}, need to be ["prototype_name" or "category"]'
                )

            # Perform mapping
            if key_to_map in self.category_mapping:
                category_name = self.category_mapping[key_to_map][0]
                category_id = int(self.category_mapping[key_to_map][1])
            else:
                category_name = "other"
                category_id = ATEK_OTHER_CATETORY_ID

        return category_name, category_id

    def _sample_points_on_bbox(
        self, bbox2d_range: np.array, num_points_on_edge: int
    ) -> torch.Tensor:
        """
        Samples points on the edges of a 2D bounding box.
        Args:
            bbox2d_range (np.array): An array containing the bounding box coordinates [xmin, xmax, ymin, ymax].
            num_points_on_edge (int): The number of points to sample along each edge of the bounding box.
        Returns:
            torch.Tensor: A tensor containing the sampled pixel coordinates.
        """
        # Create sampled pixel coordinates along each edge of the 2d bbox
        xmin, xmax, ymin, ymax = bbox2d_range
        x = torch.linspace(
            xmin, xmax, num_points_on_edge + 2
        )  # K+2 to include the corners
        y = torch.linspace(
            ymin, ymax, num_points_on_edge + 2
        )  # K+2 to include the corners

        # Create the points on the edges
        top_edge = torch.stack((x, torch.full_like(x, ymin)), dim=1)
        bottom_edge = torch.stack((x, torch.full_like(x, ymax)), dim=1)
        left_edge = torch.stack((torch.full_like(y, xmin), y), dim=1)
        right_edge = torch.stack((torch.full_like(y, xmax), y), dim=1)

        # Concatenate all the points
        pixel_coords = torch.cat((top_edge, bottom_edge, left_edge, right_edge), dim=0)

        return pixel_coords

    def _apply_transforms_to_bbox2d(
        self, camera_label: str, bbox2d_range: np.array
    ) -> torch.Tensor:
        """
        Apply the same transforms in AriaCameraProcessors to 2D bounding boxes.
        Returns a new 2dbbox that encloses the distorted box.
        """
        src_sampled_points = self._sample_points_on_bbox(
            bbox2d_range, self.conf.bbox2d_num_samples_on_edge
        )
        dst_sampled_points = self.camera_label_to_pixel_transforms[camera_label](
            src_sampled_points
        )

        # Get the new 2d bbox range
        dst_image_width, dst_image_height = self.camera_label_to_calibs[
            camera_label
        ].get_image_size()
        xmin, ymin = torch.min(dst_sampled_points, dim=0).values
        xmax, ymax = torch.max(dst_sampled_points, dim=0).values
        xmin = torch.clamp(xmin, 0, dst_image_width - 1)
        xmax = torch.clamp(xmax, 0, dst_image_width - 1)
        ymin = torch.clamp(ymin, 0, dst_image_height - 1)
        ymax = torch.clamp(ymax, 0, dst_image_height - 1)

        return torch.tensor([xmin, xmax, ymin, ymax], dtype=torch.float32)

    def get_gt_by_timestamp_ns(self, timestamp_ns: int) -> Optional[Dict]:
        """
        Retrieves the ground truth data for 2D bounding boxes by timestamp in nanoseconds.

        Args:
            timestamp_ns (int): The timestamp in nanoseconds for which to retrieve the ground truth data.

        Returns:
            Optional[Dict]: A dictionary containing the ground truth data for 2D bounding boxes if available and valid; otherwise, None.

            If not None, the returned dictionary will have the following structure:
            {
                "camera_label_1": {
                    "instance_ids": torch.Tensor (shape: [num_instances], int64)
                    "category_names": list[str],
                    "category_ids": torch.Tensor (shape: [num_instances], int64),
                    "visibility_ratios": torch.Tensor (shape: [num_instances], float32),
                    "box_ranges": torch.Tensor (shape: [num_instances, 4], float32, [xmin, xmax, ymin, ymax])
                },
                "camera_label_2": {
                    ...
                }
                ...
            }
            Each key in the outer dictionary corresponds to a camera label, while each inner dictionary contains instance information for that camera.

        Notes:
            Returns None if the data at the specified timestamp is not valid or does not meet the configured tolerances.
        """
        bbox2d_dict = {}
        for cam_label, stream_id in self.camera_label_to_stream_ids.items():

            bbox2d_with_dt = (
                self.adt_gt_provider.get_object_2d_boundingboxes_by_timestamp_ns(
                    timestamp_ns,
                    stream_id,
                )
            )
            # no valid data, skip current camera
            if (
                not bbox2d_with_dt.is_valid()
                or bbox2d_with_dt.dt_ns() > self.conf.tolerance_ns
            ):
                continue

            # initialize result dict for current camera
            num_visible_instances = len(bbox2d_with_dt.data())

            bbox2d_dict[cam_label] = {
                "instance_ids": torch.empty((num_visible_instances), dtype=torch.int64),
                "category_names": [],
                "category_ids": torch.empty((num_visible_instances), dtype=torch.int64),
                "visibility_ratios": torch.empty(
                    (num_visible_instances), dtype=torch.float32
                ),
                "box_ranges": torch.empty(
                    (num_visible_instances, 4), dtype=torch.float32
                ),
            }

            if num_visible_instances == 0:
                logger.debug(
                    f"No visible 2d bbox data for camera {cam_label} at {timestamp_ns}, skipping"
                )
                continue

            # pack 2d bbox data into the dict
            i_row = 0
            for instance_id, bbox2d_data in bbox2d_with_dt.data().items():
                # fill in instance id and category information
                cat_name, cat_id = self._obtain_obj_category_info(instance_id)
                bbox2d_dict[cam_label]["instance_ids"][i_row] = instance_id
                bbox2d_dict[cam_label]["category_names"].append(cat_name)
                bbox2d_dict[cam_label]["category_ids"][i_row] = cat_id
                bbox2d_dict[cam_label]["visibility_ratios"][
                    i_row
                ] = bbox2d_data.visibility_ratio
                bbox2d_dict[cam_label]["box_ranges"][i_row] = (
                    self._apply_transforms_to_bbox2d(cam_label, bbox2d_data.box_range)
                )
                i_row += 1
            assert (
                i_row == num_visible_instances
            ), f"camera {cam_label} filled number {i_row} != num of instances {num_visible_instances}, tensor contains initialized values, unsafe hence abort"

        # At least one camera should have valid data, or will return None
        valid_data_flag = False
        for per_cam_dict in bbox2d_dict.values():
            if len(per_cam_dict["category_names"]) > 0:
                valid_data_flag = True

        if not valid_data_flag:
            return None
        else:
            return bbox2d_dict



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-strict

import logging
import os

from typing import Any, Callable, Dict, Optional

from atek.data_loaders.atek_wds_dataloader import (
    process_wds_sample,
    select_and_remap_dict_keys,
)
from atek.data_loaders.cubercnn_model_adaptor import CubeRCNNModelAdaptor
from atek.data_preprocess.atek_data_sample import AtekDataSample

from atek.data_preprocess.sample_builders.obb_sample_builder import ObbSampleBuilder
from atek.data_preprocess.subsampling_lib.temporal_subsampler import (
    CameraTemporalSubsampler,
)
from omegaconf.omegaconf import DictConfig

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class AtekRawDataloaderAsCubercnn:
    def __init__(
        self, vrs_file: str, mps_files: Dict[str, str], conf: DictConfig
    ) -> None:
        # initialize the sample builder
        self.sample_builder = ObbSampleBuilder(
            conf=conf.processors, vrs_file=vrs_file, mps_files=mps_files, gt_files={}
        )

        # Create a subsampler
        self.subsampler = CameraTemporalSubsampler(
            vrs_file=vrs_file,
            conf=conf.camera_temporal_subsampler,
        )

        # Create a CubeRCNN model adaptor
        self.model_adaptor = CubeRCNNModelAdaptor()

    def __len__(self):
        return self.subsampler.get_total_num_samples()

    def get_timestamps_by_sample_index(self, index: int):
        return self.subsampler.get_timestamps_by_sample_index(index)

    def get_atek_sample_at_timestamp_ns(
        self, timestamp_ns: int
    ) -> Optional[AtekDataSample]:
        return self.sample_builder.get_sample_by_timestamp_ns(timestamp_ns)

    def get_model_specific_sample_at_timestamp_ns(
        self, timestamp_ns: int
    ) -> Optional[Dict]:
        atek_sample = self.sample_builder.get_sample_by_timestamp_ns(timestamp_ns)
        if atek_sample is None:
            logger.warning(
                f"Cannot retrieve valid atek sample at timestamp {timestamp_ns}"
            )
            return None

        # Flatten to dict
        atek_sample_dict = atek_sample.to_flatten_dict()

        # key remapping
        remapped_data_dict = select_and_remap_dict_keys(
            sample_dict=atek_sample_dict,
            key_mapping=self.model_adaptor.get_dict_key_mapping_all(),
        )

        # transform
        model_specific_sample_gen = self.model_adaptor.atek_to_cubercnn(
            [remapped_data_dict]
        )

        # Obtain a dict from a generator object
        model_specific_sample = next(model_specific_sample_gen)

        return model_specific_sample



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-strict

import unittest

import torch
from atek.util.file_io_utils import merge_tensors_into_dict, separate_tensors_from_dict
from atek.util.tensor_utils import check_dicts_same_w_tensors


class FileIoUtilsTest(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_separate_tensors_from_dict(self) -> None:
        input_dict = {
            "key_1": {
                "subkey_1": torch.tensor([1, 2, 3]),
                "subkey_2": torch.tensor([4, 5, 6]),
                "subkey_3": "value",
            },
            "key_2": [2, 3, 4],  # List should stay
            "key_3": {
                "key_4": {
                    "key_5": torch.tensor([7, 8]),
                    "key_6": {
                        "key_8": 1000,
                        "key_9": torch.tensor([10, 11, 12]),
                    },
                }
            },
        }

        # Test the function
        dict_wo_tensors, tensor_dict = separate_tensors_from_dict(input_dict)

        expected_dict_wo_tensors = {
            "key_1": {
                "subkey_3": "value",
            },
            "key_2": [2, 3, 4],  # List should stay
            "key_3": {
                "key_4": {
                    "key_6": {
                        "key_8": 1000,
                    }
                }
            },
        }
        expected_tensor_dict = {
            "key_1+subkey_1": torch.tensor([1, 2, 3]),
            "key_1+subkey_2": torch.tensor([4, 5, 6]),
            "key_3+key_4+key_5": torch.tensor([7, 8]),
            "key_3+key_4+key_6+key_9": torch.tensor([10, 11, 12]),
        }

        self.assertDictEqual(dict_wo_tensors, expected_dict_wo_tensors)
        self.assertEqual(set(tensor_dict.keys()), set(expected_tensor_dict.keys()))
        for key in tensor_dict:
            self.assertTrue(
                torch.allclose(tensor_dict[key], expected_tensor_dict[key], atol=0)
            )

    def test_separate_tensor_from_dict_round_trip(self) -> None:
        input_dict = {
            "key_1": {
                "subkey_1": torch.tensor([1, 2, 3]),
                "subkey_2": torch.tensor([4, 5, 6]),
                "subkey_3": "value",
            },
            "key_2": [2, 3, 4],  # List should stay
            "key_3": {
                "key_4": {
                    "key_5": torch.tensor([7, 8]),
                    "key_6": {
                        "key_8": 1000,
                        "key_9": torch.tensor([10, 11, 12]),
                    },
                }
            },
        }

        dict_wo_tensors, tensor_dict = separate_tensors_from_dict(input_dict)
        result_dict = merge_tensors_into_dict(dict_wo_tensors, tensor_dict)

        self.assertTrue(check_dicts_same_w_tensors(input_dict, result_dict))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

import numpy as np

import torch

from atek.data_preprocess.processors.obb3_gt_processor import Obb3GtProcessor
from omegaconf import OmegaConf
from projectaria_tools.core.stream_id import StreamId


# test data paths
TEST_DIR_PATH = os.path.join(os.getenv("TEST_FOLDER"))
CONFIG_PATH = os.getenv("CONFIG_PATH")
CATEGORY_MAPPING_PATH = os.getenv("CATEGORY_MAPPING_PATH")


class Obb3GtProcessorTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_get_obb3_gt_data(self) -> None:
        conf = OmegaConf.load(CONFIG_PATH)
        conf.processors.obb_gt.category_mapping_field_name = "prototype_name"

        camera_label_to_stream_ids = {
            "camera-rgb": StreamId("214-1"),
            "camera-slam-left": StreamId("1201-1"),
        }

        obb3_gt_processor = Obb3GtProcessor(
            obb3_file_path=os.path.join(TEST_DIR_PATH, "test_3d_bounding_box.csv"),
            obb3_traj_file_path=os.path.join(
                TEST_DIR_PATH, "test_3d_bounding_box_traj.csv"
            ),
            obb2_file_path=os.path.join(TEST_DIR_PATH, "test_2d_bounding_box.csv"),
            instance_json_file_path=os.path.join(TEST_DIR_PATH, "test_instances.json"),
            category_mapping_file_path=CATEGORY_MAPPING_PATH,
            camera_label_to_stream_ids=camera_label_to_stream_ids,
            conf=conf.processors.obb_gt,
        )

        # Test for valid query
        queried_obb3_data = obb3_gt_processor.get_gt_by_timestamp_ns(
            timestamp_ns=87551170910700
        )
        rgb_visible_instances = queried_obb3_data["camera-rgb"]

        # Check bbox gt tensor shapes
        gt_num_instances = 91
        self.assertEqual(
            rgb_visible_instances["category_ids"].shape,
            torch.Size([gt_num_instances]),
        )
        self.assertEqual(len(rgb_visible_instances["category_names"]), gt_num_instances)
        self.assertEqual(
            rgb_visible_instances["object_dimensions"].shape,
            torch.Size([gt_num_instances, 3]),
        )
        self.assertEqual(
            rgb_visible_instances["ts_world_object"].shape,
            torch.Size([gt_num_instances, 3, 4]),
        )

        # Check the content of a specific instance 3d bbox
        instance_id = 4213328128795167
        ind = torch.where(rgb_visible_instances["instance_ids"] == instance_id)[0]
        self.assertTrue(len(ind) == 1)
        ind = ind.item()

        gt_cat_name = "other"
        gt_cat_id = 0
        gt_obj_dim = torch.tensor([0.35102427005, 0.17546123076, 0.24489420652])

        self.assertEqual(gt_cat_name, rgb_visible_instances["category_names"][ind])
        self.assertEqual(gt_cat_id, rgb_visible_instances["category_ids"][ind].item())
        self.assertTrue(
            torch.allclose(gt_obj_dim, rgb_visible_instances["object_dimensions"][ind])
        )
